#include <iostream>
#include <glut.h>
#include <OpenGL/gl.h>
#include <stdint.h>
#include <list>
#include "glm/glm.hpp"
#include "glm/ext.hpp"

struct Vertex{
    glm::vec3 position;
    glm::vec3 normal;
    glm::vec2 texcoord;
    glm::vec4 joint;
    glm::vec4 weight;
};

struct Vbo{
    bool loadFromData(void *verticesData, uint64_t vertexSize, void *indicesData, uint64_t indicesSize){
        glGenBuffers(1, &vertex);
        glBindBuffer(GL_ARRAY_BUFFER, vertex);
        glBufferData(GL_ARRAY_BUFFER, vertexSize, verticesData, GL_STATIC_DRAW);
        glGenBuffers(1, &index);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indicesSize, indicesData, GL_STATIC_DRAW);
        return true;
    }

    void release(){
        glDeleteBuffers(1, &vertex);
        glDeleteBuffers(1, &index);
    }

    void bind(){
        glBindBuffer(GL_ARRAY_BUFFER, vertex);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index);
    }

    GLuint vertex;
    GLuint index;
};

struct Attribute{
    Attribute(GLint handle, uint32_t size, uint64_t offset)
        : handle(handle), size(size), offset(offset) {}
    GLint handle;
    uint32_t size;
    uint64_t offset;
};

struct Shader{
    bool loadFromSources(const std::string &vertexSource, const std::string &fragmentSource){
        vertex = glCreateShader(GL_VERTEX_SHADER);
        fragment = glCreateShader(GL_FRAGMENT_SHADER);
        const char *vertexSrc = vertexSource.c_str();
        const char *fragmentSrc = fragmentSource.c_str();
        int vertexSrcLen = strlen(vertexSrc);
        int fragmentSrcLen = strlen(fragmentSrc);
        glShaderSource(vertex, 1, &vertexSrc, &vertexSrcLen);
        glShaderSource(fragment, 1, &fragmentSrc, &fragmentSrcLen);
        glCompileShader(vertex);
        glCompileShader(fragment);
        program = glCreateProgram();
        glAttachShader(program, vertex);
        glAttachShader(program, fragment);
        glLinkProgram(program);
        return true;
    }

    void release(){
        glDeleteShader(vertex);
        glDeleteShader(fragment);
        glDeleteProgram(program);
    }

    GLuint getAttributeHandle(const std::string &name){
        return glGetAttribLocation(program, name.c_str());
    }

    GLuint getUniformHandle(const std::string &name){
        return glGetUniformLocation(program, name.c_str());
    }

    void beginDraw(){
        bind();
        for (std::list<Attribute>::iterator it = attributes.begin(), end = attributes.end(); it != end; ++it){
            glEnableVertexAttribArray(it->handle);
            glVertexAttribPointer(it->handle, it->size, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(it->offset));
        }
    }

    void endDraw(){
        for (std::list<Attribute>::iterator it = attributes.begin(), end = attributes.end(); it != end; ++it)
            glDisableVertexAttribArray(it->handle);
    }

    void bind(){
        glUseProgram(program);
    }

    GLuint program;
    GLuint vertex;
    GLuint fragment;
    GLint mvp;
    GLint texture;
    std::list<Attribute> attributes;
};

struct Texture{
    bool loadFromData(void *bitmapData, uint16_t width, uint16_t height, uint8_t bpp){
        glGenTextures(1, &handle);
        glBindTexture(GL_TEXTURE_2D, handle);
        GLenum type = GL_RGBA;
        if (bpp == 24) type = GL_RGB;
        glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, bitmapData);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glGenerateMipmap(GL_TEXTURE_2D);
        return true;
    }

    void bind(uint8_t layer){
        glBindTexture(GL_TEXTURE_2D, handle);
        glActiveTexture(GL_TEXTURE0 + layer);
    }

    void release(){
        glDeleteTextures(1, &handle);
    }

    GLuint handle;
};

struct Color{
    Color(uint8_t r = 255, uint8_t g = 255, uint8_t b = 255, uint8_t a = 255)
        : r(r), g(g), b(b), a(a) {}
    uint8_t r, g, b, a;
};

// variables
float rotation = 0;
Vbo quad;
Shader shader;
Texture texture;
const char *vertexSrc =
        "attribute vec4 position;\n"
        "attribute vec2 texcoord;\n"
        "uniform mat4 mvp;\n"
        "varying vec2 v_texcoord;\n"
        "void main(){\n"
        "   gl_Position = mvp * position;\n"
        "   v_texcoord = texcoord;\n"
        "}\n"
        "";
const char *fragmentSrc =
        "uniform sampler2D texture;\n"
        "varying vec2 v_texcoord;\n"
        "void main(){\n"
        "   gl_FragColor = texture2D(texture, v_texcoord);\n"
        "}\n"
        "";

void init(){
    glEnable(GL_DEPTH_TEST);

    // init quad buffers
    Vertex quadVertices[4];
    quadVertices[0].position = glm::vec3(-1, -1, 0); quadVertices[0].normal = glm::vec3(0, 0, -1); quadVertices[0].texcoord = glm::vec2(0, 0);
    quadVertices[1].position = glm::vec3(-1, 1, 0); quadVertices[1].normal = glm::vec3(0, 0, -1); quadVertices[1].texcoord = glm::vec2(0, 1);
    quadVertices[2].position = glm::vec3(1, 1, 0); quadVertices[2].normal = glm::vec3(0, 0, -1); quadVertices[2].texcoord = glm::vec2(1, 1);
    quadVertices[3].position = glm::vec3(1, -1, 0); quadVertices[3].normal = glm::vec3(0, 0, -1); quadVertices[3].texcoord = glm::vec2(1, 0);
    uint16_t quadIndices[6] = {0,1,2,0,2,3};

    // create quad buffers
    quad.loadFromData(&quadVertices[0], 4 * sizeof(Vertex), &quadIndices[0], 6 * sizeof(uint16_t));

    // create shader
    shader.loadFromSources(vertexSrc, fragmentSrc);

    // init shader
    shader.attributes.push_back(Attribute(shader.getAttributeHandle("position"), 3, 0));
    shader.attributes.push_back(Attribute(shader.getAttributeHandle("texcoord"), 2, 24));
    shader.mvp = shader.getUniformHandle("mvp");
    shader.texture = shader.getUniformHandle("texture");

    // create texture
    Color bitmap[32 * 32];
    for (uint y = 0; y < 32; ++y)
        for (uint x = 0; x < 32; ++x){
            bitmap[y * 32 + x] = Color(float(x) / 32.0f * 255, float(y) / 32.0f * 255, 1.0 * 255, 1.0 * 255);
        }
    texture.loadFromData(bitmap, 32, 32, 32);
}

void render(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // bind vbo
    quad.bind();

    // prepare attributes
    shader.beginDraw();

    // prepare shader
    glm::mat4 mvp = glm::perspective(45.0f, 800.0f / 480.0f, 1.0f, 100.0f) * glm::lookAt(glm::vec3(0, 0, -5), glm::vec3(0.0), glm::vec3(0, 1, 0)) * glm::eulerAngleYXZ(rotation+=0.01, 0.0f, 0.0f);
    glUniformMatrix4fv(shader.mvp, 1, GL_FALSE, &mvp[0][0]);
    texture.bind(0);
    glUniform1i(shader.texture, 0);
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);

    shader.endDraw();

    glutSwapBuffers();
}

void reshape(int w, int h){
    glViewport(0, 0, w, h);
}

void cleanup(){
    texture.release();
    shader.release();
    quad.release();
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(800, 480);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("Test");

    init();

    std::cout << "Vertex size " << sizeof(Vertex) << " bytes" << std::endl;
    std::cout << "Quad: vertex " << quad.vertex << " index " << quad.index << std::endl;

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutIdleFunc(render);

    glutMainLoop();

    cleanup();
    return 0;
}

